/* coreography, a tool for examining ELF based memory images.
 * Copyright (C) 2003  John Ventura ventura@engination.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <elf.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/procfs.h>
#include <sys/ptrace.h>
#include "cmd.h"
#include "addresses.h"

void *getauxinfo(uint8_t * buf, int len, int type);
int regdumppid(int pid, int cmdval, FILE * desc);
uint8_t *getprocseg(int pid, uint32_t addr, int len);

struct reg32 {
    uint32_t ebx;
    uint32_t ecx;
    uint32_t edx;
    uint32_t esi;
    uint32_t edi;
    uint32_t ebp;
    uint32_t eax;
    uint32_t xds;
    uint32_t xes;
    uint32_t xfs;
    uint32_t xgs;
    uint32_t orig_eax;
    uint32_t eip;
    uint32_t xcs;
    uint32_t eflags;
    uint32_t oldesp;
    uint32_t oldss;

};

/* This function reads a core dump traces the frames.
 */
int frame(uint8_t * buf, int len, int cmdval, int usepid, int pid,
	  FILE * desc)
{
    int i;
    int aux_offset = 0;
    int aux_size = 0;
    int stackoffset = 0;
    uint32_t stackstart = 0;
    uint32_t *x = NULL;
    uint32_t eip;

    Elf32_Ehdr *elfheader;
    Elf32_Phdr *programheader;
    struct reg32 *registers = NULL;
    struct elf_prstatus *prstatus;

    if (usepid) {
	regdumppid(pid, cmdval, desc);
	return (0);
    }

    elfheader = (Elf32_Ehdr *) buf;
    programheader = (Elf32_Phdr *) (buf + elfheader->e_phoff);

    if (elfheader->e_type != ET_CORE) {
	fprintf(stderr, "this is not a core dump\n");
	exit(1);
    }

    /* find auxiliary data */
    for (i = 0; i < elfheader->e_phnum; i++) {
	if (programheader[i].p_type == PT_NOTE) {
	    aux_offset = programheader[i].p_offset;
	    aux_size = programheader[i].p_filesz;
	}
	if ((programheader[i].p_vaddr + programheader[i].p_filesz) ==
	    END_STACK_LINUX)
	    stackoffset = programheader[i].p_offset;
	stackstart = programheader[i].p_vaddr;
    }

    prstatus =
	(struct elf_prstatus *) getauxinfo(buf + aux_offset, aux_size,
					   NT_PRSTATUS);
    if (prstatus == NULL) {
	fprintf(stderr, "register information not saved\n");
	exit(1);
    }
    registers = (struct reg32 *) (prstatus->pr_reg);
    if (cmdval == CMD_REGDUMP) {
	fprintf(desc, "EBX =\t\t%08x\n", registers->ebx);
	fprintf(desc, "ECX =\t\t%08x\n", registers->ecx);
	fprintf(desc, "EDX =\t\t%08x\n", registers->edx);
	fprintf(desc, "ESI =\t\t%08x\n", registers->esi);
	fprintf(desc, "EDI =\t\t%08x\n", registers->edi);
	fprintf(desc, "EBP =\t\t%08x\n", registers->ebp);
	fprintf(desc, "EAX =\t\t%08x\n", registers->eax);
	fprintf(desc, "DS =\t\t%08x\n", registers->xds);
	fprintf(desc, "ES =\t\t%08x\n", registers->xes);
	fprintf(desc, "ORIG_EAX =\t%08x\n", registers->orig_eax);
	fprintf(desc, "EIP =\t\t%08x\n", registers->eip);
	fprintf(desc, "CS =\t\t%08x\n", registers->xcs);
	fprintf(desc, "EFLAGS =\t%08x\n", registers->eflags);
	fprintf(desc, "OLDESP =\t%08x\n", registers->oldesp);
	fprintf(desc, "OLDSS =\t\t%08x\n", registers->oldss);
    } else {

	buf += stackoffset;

	x = (uint32_t *) ((registers->oldesp - stackstart) + (buf));
	eip = *x;		/* initial valu for frame 1 EIP */

	/* Frame 0 is the current frame.  We starting at frame 1, because
	 * if you want info on frame 0, regdump will give you much more.
	 */
	for (i = 1; registers->ebp != 0; i++) {

	    x = (uint32_t *) ((registers->ebp - stackstart) + (buf));
	    fprintf(desc, "Stack frame %i at %08x\n", i, registers->ebp);
	    fprintf(desc, "\tsaved EIP = %08x\n", *(x + 1));
	    fprintf(desc, "\tEIP = %08x\n", eip);
	    eip = *(x + 1);
	    registers->ebp = *x;
	}
    }
    return (0);
}

int regdumppid(int pid, int cmdval, FILE * desc)
{
    int wait_val;
    struct user_regs_struct regs;

    if (ptrace(PTRACE_ATTACH, pid, NULL, NULL) < 0) {
	fprintf(stderr, "can't attach to process %i\n", pid);
	exit(1);
    }
    wait(&wait_val);

    if (ptrace(PTRACE_GETREGS, pid, NULL, &regs) < 0) {
	fprintf(stderr, "can't get register information from process %i\n",
		pid);
	exit(1);
    }
    ptrace(PTRACE_DETACH, pid, NULL, NULL);

    if (cmdval == CMD_REGDUMP) {
	fprintf(desc, "EBX =\t\t%08x\n", (uint32_t) regs.ebx);
	fprintf(desc, "ECX =\t\t%08x\n", (uint32_t) regs.ecx);
	fprintf(desc, "EDX =\t\t%08x\n", (uint32_t) regs.edx);
	fprintf(desc, "ESI =\t\t%08x\n", (uint32_t) regs.esi);
	fprintf(desc, "EDI =\t\t%08x\n", (uint32_t) regs.edi);
	fprintf(desc, "EBP =\t\t%08x\n", (uint32_t) regs.ebp);
	fprintf(desc, "EAX =\t\t%08x\n", (uint32_t) regs.eax);
	fprintf(desc, "DS =\t\t%08x\n", (uint32_t) regs.xds);
	fprintf(desc, "ES =\t\t%08x\n", (uint32_t) regs.xes);
	fprintf(desc, "ORIG_EAX =\t%08x\n", (uint32_t) regs.orig_eax);
	fprintf(desc, "EIP =\t\t%08x\n", (uint32_t) regs.eip);
	fprintf(desc, "CS =\t\t%08x\n", (uint32_t) regs.xcs);
	fprintf(desc, "ESP =\t\t%08x\n", (uint32_t) regs.esp);
	fprintf(desc, "EFLAGS =\t%08x\n", (uint32_t) regs.eflags);
	fprintf(desc, "OLDSS =\t\t%08x\n", (uint32_t) regs.xss);
    }
    return (0);
}
