/* coreography, a tool for examining ELF based memory images.
 * Copyright (C) 2003  John Ventura ventura@engination.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <elf.h>
#include <time.h>
#include <sys/types.h>
#include <sys/procfs.h>
#include "addresses.h"
#include "elfdefs.h"
#include "sig.h"

#define NELEMS(array) (sizeof(array) / sizeof(array[0]))
#define isnotchar(c)((c < 0x20) || (c > 0x7e))

int getinternaloffset(uint8_t * buf);
void *getauxinfo(uint8_t * buf, int len, int type);
int checktextsanity(uint8_t * buf, int len);
uint8_t *getprocseg(int pid, uint32_t addr, int len);

/* This function gives readable output that is generated after parsing
 * an ELF file.  *buf is the address to the buffer of len bytes.
 * If internal is true, it assumes the ELF is a core dump, and 
 * tries to read the internal headers/
 */

int info(uint8_t * buf, int len, int internal, int binary, int usepid,
	 int pid, FILE * desc)
{
    int i;
    int n;
    int textsanity = 0;
    char *strtab = NULL;
    char delim = '\n';
    struct elf_prstatus *prstatus;
    struct elf_prpsinfo *prpsinfo;
    Elf32_Ehdr *elfheader;
    Elf32_Phdr *programheader;
    Elf32_Shdr *sectionheader;

    /* if we don't want human readable output, use comma instead
     * of newline character to seperate lines.
     */
    if (binary)
	delim = ',';

    if (usepid)
	buf = (uint8_t *) getprocseg(pid, START_TEXT_LINUX, 0x1000);

    elfheader = (Elf32_Ehdr *) buf;
    programheader = (Elf32_Phdr *) (buf + elfheader->e_phoff);

    /* if we are looking at the external headers, and it really is
     * a core dump, then parse auxiliary info.
     */
    if ((!internal) && (elfheader->e_type == ET_CORE)) {
	for (i = 0; i < elfheader->e_phnum; i++) {
	    if (programheader[i].p_type == PT_NOTE) {
		prstatus = (struct elf_prstatus *)
		    getauxinfo(buf + programheader[i].p_offset,
			       programheader[i].p_filesz, NT_PRSTATUS);
		prpsinfo = (struct elf_prpsinfo *)
		    getauxinfo(buf + programheader[i].p_offset,
			       programheader[i].p_filesz, NT_PRPSINFO);
		if (prstatus) {
		    fprintf(desc, "Signal =\t\t%i\t",
			    prstatus->pr_info.si_signo);
		    for (n = 0; n < NELEMS(sigtab); n++) {
			if (prstatus->pr_info.si_signo == sigtab[n].signal)
			    fprintf(desc, "%s \"%s\"", sigtab[n].signame,
				    sigtab[n].description);
		    }
		    fprintf(desc, "%c", delim);
		    fprintf(desc, "Current Signal =\t%i%c",
			    prstatus->pr_cursig, delim);
		    fprintf(desc, "Extra Code =\t\t%i%c",
			    prstatus->pr_info.si_code, delim);
		    fprintf(desc, "Error Number =\t\t%i%c",
			    prstatus->pr_info.si_errno, delim);
		    if (prstatus->pr_fpvalid)
			fprintf(desc, "coprocessor used");
		    else
			fprintf(desc, "coprocessor not used");
		    fprintf(desc, "%c", delim);
		}
		if (prpsinfo != NULL) {
		    fprintf(desc, "Process state =\t\t%i%c",
			    prpsinfo->pr_state, delim);
		    fprintf(desc, "Process state char =\t\"%c\"%c",
			    prpsinfo->pr_sname, delim);
		    fprintf(desc, "Zombie Value =\t\t%i%c",
			    prpsinfo->pr_zomb, delim);
		    fprintf(desc, "Nice Value =\t\t%i%c",
			    prpsinfo->pr_nice, delim);
		    fprintf(desc, "UID =\t\t\t%i%c", prpsinfo->pr_uid,
			    delim);
		    fprintf(desc, "GID =\t\t\t%i%c", prpsinfo->pr_gid,
			    delim);
		    fprintf(desc, "PID =\t\t\t%i%c", prpsinfo->pr_pid,
			    delim);
		    fprintf(desc, "PPID =\t\t\t%i%c", prpsinfo->pr_ppid,
			    delim);
		    fprintf(desc, "PGRP =\t\t\t%i%c", prpsinfo->pr_pgrp,
			    delim);
		    fprintf(desc, "SID =\t\t\t%i%c", prpsinfo->pr_sid,
			    delim);
		    fprintf(desc, "filename =\t\t%s%c\n",
			    prpsinfo->pr_fname, delim);
		}
		break;
	    }
	}
    }

    fprintf(desc, "\nELF Headers:\n");
    fprintf(desc, "e_type =\t%08x\t", elfheader->e_type);
    for (i = 0; i < NELEMS(etype); i++) {
	if (elfheader->e_type == etype[i].value)
	    fprintf(desc, "%s%c", etype[i].str, delim);
    }

    fprintf(desc, "e_machine =\t%08x\t", elfheader->e_machine);
    for (i = 0; i < NELEMS(emachine); i++) {
	if (elfheader->e_machine == emachine[i].value)
	    fprintf(desc, "%s%c", emachine[i].str, delim);
    }

    fprintf(desc, "e_version =\t%08x%c", elfheader->e_version, delim);
    fprintf(desc, "e_phoff =\t%08x%c", elfheader->e_phoff, delim);
    fprintf(desc, "e_shoff =\t%08x%c", elfheader->e_shoff, delim);
    fprintf(desc, "e_flags =\t%08x%c", elfheader->e_flags, delim);
    fprintf(desc, "e_ehsize =\t%04x%c", elfheader->e_ehsize, delim);
    fprintf(desc, "e_phentssize =\t%04x%c", elfheader->e_phentsize, delim);
    fprintf(desc, "e_phnum =\t%04x%c", elfheader->e_phnum, delim);
    fprintf(desc, "e_shentsize =\t%04x%c", elfheader->e_shentsize, delim);
    fprintf(desc, "e_shnum =\t%04x%c", elfheader->e_shnum, delim);
    fprintf(desc, "e_shstrndx =\t%04x%c\n", elfheader->e_shstrndx, delim);

    fprintf(desc, "Program headers:\n");
    for (i = 0; i < elfheader->e_phnum; i++) {
	fprintf(desc, "p_type =\t%08x\t", programheader[i].p_type);
	for (n = 0; n < NELEMS(ptype); n++) {
	    if (programheader[i].p_type == ptype[n].value)
		fprintf(desc, "%s \"%s\"", ptype[n].tagstr,
			ptype[n].description);
	}
	fprintf(desc, "%c", delim);
	fprintf(desc, "p_offset =\t%08x%c", programheader[i].p_offset,
		delim);
	fprintf(desc, "p_vaddr =\t%08x%c", programheader[i].p_vaddr,
		delim);
	fprintf(desc, "p_paddr =\t%08x%c", programheader[i].p_paddr,
		delim);
	fprintf(desc, "p_filesz =\t%08x%c", programheader[i].p_filesz,
		delim);
	fprintf(desc, "p_memsz =\t%08x%c", programheader[i].p_memsz,
		delim);
	fprintf(desc, "p_flags =\t%08x\t", programheader[i].p_flags);
	if (programheader[i].p_flags & PF_R)
	    fprintf(desc, "READ ");
	if (programheader[i].p_flags & PF_W)
	    fprintf(desc, "WRITE ");
	if (programheader[i].p_flags & PF_X)
	    fprintf(desc, "EXEC ");
	if (programheader[i].p_flags & PF_MASKOS)
	    fprintf(desc, "OS ");
	if (programheader[i].p_flags & PF_MASKPROC)
	    fprintf(desc, "PROC ");
	fprintf(desc, "%c", delim);
	fprintf(desc, "p_align =\t%08x%c\n", programheader[i].p_align,
		delim);
    }

    /* This bit of code dumps the section header info.
     * If we are looking at "internal" headers, don't dump anything,
     * because section headers won't be there.  Also,make sure
     * that there really is room in the buffer for the section headers
     * as described in the Elf32_Ehdr.
     */

    if ((!internal) && (!usepid)
	&& (elfheader->e_shoff +
	    (elfheader->e_shnum * sizeof(Elf32_Shdr)) <= len)) {
	sectionheader = (Elf32_Shdr *) (buf + elfheader->e_shoff);
	/* make sure string table index is actually in buffer and
	 * has valid data 
	 */
	if (sectionheader[elfheader->e_shstrndx].sh_offset < len) {
	    textsanity = checktextsanity((buf +
					  sectionheader[elfheader->
							e_shstrndx].
					  sh_offset),
					 sectionheader[elfheader->
						       e_shstrndx].
					 sh_size);
	    strtab =
		(char *) (buf +
			  sectionheader[elfheader->e_shstrndx].sh_offset);
	}
	for (i = 0; i < elfheader->e_shnum; i++) {
	    fprintf(desc, "sh_name =\t%08x\t", sectionheader[i].sh_name);
	    if ((textsanity) &&
		(sectionheader[i].sh_name <
		 sectionheader[elfheader->e_shstrndx].sh_size))
		fprintf(desc, "%s", (strtab + sectionheader[i].sh_name));
	    fprintf(desc, "%c", delim);
	    fprintf(desc, "sh_type =\t%08x\t", sectionheader[i].sh_type);
	    for (n = 0; n < NELEMS(shtype); n++) {
		if (sectionheader[i].sh_type == shtype[n].value)
		    fprintf(desc, "%s \"%s\"%c", shtype[n].tagstr,
			    shtype[n].description, delim);
	    }

	    fprintf(desc, "sh_flags =\t%08x\t", sectionheader[i].sh_flags);
	    if (sectionheader[i].sh_flags & SHF_WRITE)
		fprintf(desc, "WRITE ");
	    if (sectionheader[i].sh_flags & SHF_ALLOC)
		fprintf(desc, "OCCUPIES MEMORY ");
	    if (sectionheader[i].sh_flags & SHF_EXECINSTR)
		fprintf(desc, "EXEC ");
	    fprintf(desc, "%c", delim);
	    fprintf(desc, "sh_addr =\t%08x%c", sectionheader[i].sh_addr,
		    delim);
	    fprintf(desc, "sh_offset =\t%08x%c",
		    sectionheader[i].sh_offset, delim);
	    fprintf(desc, "sh_size =\t%08x%c", sectionheader[i].sh_size,
		    delim);
	    fprintf(desc, "sh_link =\t%08x%c", sectionheader[i].sh_link,
		    delim);
	    fprintf(desc, "sh_info =\t%08x%c", sectionheader[i].sh_info,
		    delim);
	    fprintf(desc, "sh_addralign =\t%08x%c",
		    sectionheader[i].sh_addralign, delim);
	    fprintf(desc, "sh_entsize =\t%08x%c\n",
		    sectionheader[i].sh_entsize, delim);
	}
    }

    return (0);
}

/* if the chunk of text has only printable characters, return true
 */
int checktextsanity(uint8_t * buf, int len)
{
    int i;
    for (i = 0; i < len; i++) {
	if (isnotchar(buf[i]) && buf[i] != 0)
	    return (0);
    }
    return (1);
}
