/* coreography, a tool for examining ELF based memory images.
 * Copyright (C) 2003  John Ventura ventura@engination.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <elf.h>
#include <getopt.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "cmd.h"

#define QUERYBUFLEN 1024

int argdump(uint8_t * buf, int len, int binary, int usepid, int pid,
	    FILE * desc);
int regdump(uint8_t * buf, int len, FILE * desc);
int info(uint8_t * buf, int len, int internal, int binary, int usepid,
	 int pid, FILE * desc);
int frame(uint8_t * buf, int len, int cmdval, int usepid, int pid,
	  FILE * desc);
int getinternaloffset(uint8_t * buf);
int selectbytag(uint8_t * buf, int len, int segmentmask, int binary,
		int cmdval, uint8_t * query, int querylen, int usepid,
		int pid, FILE * desc);
int selectbyaddr(uint8_t * buf, int buflen, uint32_t address, uint32_t len,
		 int binary, int cmdval, uint8_t * query, int querylen,
		 int usepid, int pid, FILE * desc);
int selectbyname(uint8_t * buf, int buflen, uint32_t address, uint32_t len,
		 int binary, int cmdval, uint8_t * query, int querylen,
		 char *secname, FILE * desc);

int usage(char *progname)
{
    fprintf(stderr, "%s: <path to core> command\n", progname);
    fprintf(stderr, "Commands include:\n");
    fprintf(stderr, "\tstrings\t-- shows printable ASCII strings\n");
    fprintf(stderr, "\tsegdump\t-- dumps particular segments\n");
    fprintf(stderr, "\tregdump\t-- dumps registers at time off crash\n");
    fprintf(stderr, "\targdump\t-- dumps command line\n");
    fprintf(stderr, "\tinfo\t-- shows lots of info from headers\n");
    fprintf(stderr, "\tframe\t-- traces the frames\n");
    fprintf(stderr,
	    "\tsearch\t-- search memory image for specified data\n");
    fprintf(stderr, "\nFlags include:\n");
    fprintf(stderr, "\t-h\tPrints this\n");
    fprintf(stderr, "\t-b\t\"Binary\" instead of \"printable\" output\n");
    fprintf(stderr, "\t-o\tSpecifies an output file\n");
    fprintf(stderr, "\t-i\tSpecify internal headers\n");
    fprintf(stderr, "\t-a\tSpecify an ASCII string to search for\n");
    fprintf(stderr, "\t-x\tSpecify series of bytes to search for\n");
    fprintf(stderr, "\t-m\tLimit command to an area of memory\n");
    fprintf(stderr, "\t-S\tLimit command to named section\n");
    fprintf(stderr,
	    "\t-s\tSpecifies particular memory locations including\n");
    fprintf(stderr, "\t\taux\t-- auxiliary core file data\n");
    fprintf(stderr, "\t\ttext\t-- text segment of process\n");
    fprintf(stderr, "\t\tstack\t-- the stack\n");
    fprintf(stderr, "\t\tmmap\t-- mmaped files\n");
    fprintf(stderr, "\t\tall\t-- everything, which is the default\n");
    exit(1);
    return (0);
}

uint8_t *transquery(char *str)
{
    int i;
    char buf[3];
    uint8_t *query;

    if (strlen(str) & 0x01)	/* if it's odd numbered, truncate it */
	str[strlen(str) - 1] = 0x00;
    if (strlen(str) >= (QUERYBUFLEN * 2))	/* if it's too big, truncate it */
	str[(QUERYBUFLEN * 2)] = 0x00;
    query = (uint8_t *) malloc(strlen(str) / 2);
    for (i = 0; *str; i++) {
	strncpy(buf, str, 2);
	*(query + i) = (uint8_t) strtoul(buf, NULL, 16);
	str += 2;
    }
    return (query);
}

int main(int argc, char *argv[])
{
    int i;
    int cmdval;
    int segmentmask = 0;
    int binary = 0;
    int internal = 0;
    int internaloffset = 0;
    int source = 0;
    int rval = 0;
    char c;
    char *p, *q;
    char *corepath;
    char *outfile = NULL;
    char *cmd = NULL;
    char *lenstr;
    char *secname = NULL;
    char segments[1024];
    uint8_t *filebuf = NULL;
    uint8_t *query = NULL;
    uint32_t memaddr = 0;
    uint32_t memlen = 0;
    int querylen = 0;
    int addrselect = 0;
    int nameselect = 0;
    int usepid = 0;
    int pid = 0;
    Elf32_Ehdr *elfheader;
    struct stat fileinfo;
    FILE *output;

    while ((c = getopt(argc, argv, "ihbs:o:a:x:m:S:p")) != EOF) {
	switch (c) {
	case 'h':		/* show help info */
	    usage(argv[0]);
	    break;
	case 's':		/* chooses segments */
	    segmentmask = 0;
	    strncpy(segments, optarg, sizeof(segments));
	    for (p = segments; (q = strtok(p, ",")); p = NULL) {
		for (i = 0; segmenttab[i].value != 0; i++) {
		    if (!strcmp(segmenttab[i].name, q))
			segmentmask |= segmenttab[i].value;
		}
	    }
	    break;
	case 'b':		/* Choose non-printable output */
	    binary++;
	    break;
	case 'o':		/* choose output file */
	    outfile = optarg;
	    break;
	case 'i':		/* choose internal headers */
	    internal++;
	    break;
	case 'a':		/* search for an ASCII string */
	    query = (uint8_t *) optarg;
	    querylen = strlen(optarg);
	    break;
	case 'x':		/* search for bytes */
	    query = (uint8_t *) transquery(optarg);
	    querylen = strlen(optarg) / 2;
	    break;
	case 'm':		/* specify a chunk of memory */
	    addrselect++;
	    memaddr = strtoul(optarg, NULL, 16);
	    lenstr = strchr(optarg, ':');
	    if (lenstr != NULL)
		memlen = strtoul((lenstr + 1), NULL, 16);
	    break;
	case 'S':
	    secname = (uint8_t *) optarg;
	    nameselect++;
	    break;
	case 'p':
	    usepid++;
	    break;
	default:
	    usage(argv[0]);
	}
    }

    /* If the user specifies a file for output, then open 
     * a desciptor to it.  If no output file is given, send
     * output to standard outpuut
     */
    if (outfile != NULL) {
	output = fopen(outfile, "w+");
	if (output == NULL) {
	    fprintf(stderr, "can't open %s for writing\n", outfile);
	    exit(1);
	}
    } else
	output = stdout;

    /* Get the path to the core file */
    if (usepid)
	pid = atoi(argv[optind]);
    else {
	corepath = argv[optind];
	if (corepath == NULL) {
	    fprintf(stderr, "I NEED A PATH TO THE CORE FILE!\n");
	    exit(1);
	}

	source = open(corepath, O_RDONLY);
	if (!source) {
	    fprintf(stderr, "can't open %s\n", corepath);
	    exit(1);
	}

	/* Allocate memory large enough for core file */
	fstat(source, &fileinfo);
	filebuf = (uint8_t *) malloc(fileinfo.st_size);
	if (filebuf == NULL) {
	    fprintf(stderr, "can't read %s\n", corepath);
	    exit(1);
	}
	/* Read core file into memory */
	read(source, filebuf, fileinfo.st_size);

	/* Get the location of the "internal" headers.  These should
	 * be the same as the origiinal ELF and Program headers as the
	 * original executable that made the core file.  If we are
	 * examining the contents of these headers, then we will want to
	 * incriment the pointer to the file buffer by this much.
	 */
	if (internal)
	    internaloffset = getinternaloffset(filebuf);
	elfheader = (Elf32_Ehdr *) filebuf;
	if (memcmp(elfheader->e_ident, ELFMAG, 4) != 0) {
	    fprintf(stderr, "%s is not the right kind of file\n",
		    corepath);
	    exit(1);
	}
	/* Right now, this program pretty much just reads core files from
	 * 32-Bit Pentium Linux systems.  If this ever goes beyond a prototype
	 * the code to determine how we read files from other systems should
	 * probably go here.
	 */

	/* This part reads the user specified command, and throws input to
	 * the proper function.
	 */

    }
    cmd = argv[optind + 1];
    cmdval = 0;
    for (i = 0; cmdtab[i].value != CMD_NULL; i++) {
	if (cmd == NULL) {
	    cmdval = CMD_NULL;
	    break;
	}
	if (!strcmp(cmdtab[i].cmdname, cmd))
	    cmdval = cmdtab[i].value;
    }

    switch (cmdval) {
    case CMD_STRINGS:
	if (addrselect)
	    rval =
		selectbyaddr(filebuf + internaloffset,
			     fileinfo.st_size, memaddr, memlen, binary,
			     cmdval, NULL, 0, usepid, pid, output);
	else if (nameselect)
	    rval =
		selectbyname(filebuf + internaloffset,
			     fileinfo.st_size, memaddr, memlen, binary,
			     cmdval, query, 0, secname, output);
	else
	    rval =
		selectbytag(filebuf + internaloffset, fileinfo.st_size,
			    segmentmask, binary, cmdval, NULL, 0,
			    usepid, pid, output);
	break;
    case CMD_SEGDUMP:
	if (addrselect)
	    rval =
		selectbyaddr(filebuf + internaloffset,
			     fileinfo.st_size, memaddr, memlen, binary,
			     cmdval, NULL, 0, usepid, pid, output);
	else if (nameselect)
	    rval =
		selectbyname(filebuf + internaloffset,
			     fileinfo.st_size, memaddr, memlen, binary,
			     cmdval, query, 0, secname, output);
	else
	    rval =
		selectbytag(filebuf + internaloffset, fileinfo.st_size,
			    segmentmask, binary, cmdval, NULL, 0,
			    usepid, pid, output);
	break;
    case CMD_SEARCH:
	if (addrselect)
	    rval =
		selectbyaddr(filebuf + internaloffset,
			     fileinfo.st_size, memaddr, memlen, binary,
			     cmdval, query, querylen, usepid, pid, output);
	else if (nameselect)
	    rval =
		selectbyname(filebuf + internaloffset,
			     fileinfo.st_size, memaddr, memlen, binary,
			     cmdval, query, 0, secname, output);
	else
	    rval =
		selectbytag(filebuf + internaloffset, fileinfo.st_size,
			    segmentmask, binary, cmdval, query,
			    querylen, usepid, pid, output);
	break;
    case CMD_NULL:
	fprintf(stderr, "what exactly do you want me to do?\n");
	exit(1);
	break;
    case CMD_ARGDUMP:
	if (internal) {
	    fprintf(stderr, "not possible with internal headers\n");
	    exit(1);
	}
	argdump(filebuf, fileinfo.st_size, binary, usepid, pid, output);
	break;
    case CMD_REGDUMP:
	if (internal) {
	    fprintf(stderr, "not possible with internal headers\n");
	    exit(1);
	}
	frame(filebuf, fileinfo.st_size, cmdval, usepid, pid, output);
	break;
    case CMD_INFO:
	info(filebuf + internaloffset, fileinfo.st_size, internal,
	     binary, usepid, pid, output);
	break;
    case CMD_FRAME:
	frame(filebuf, fileinfo.st_size, cmdval, usepid, pid, output);
	break;
    default:
    }
    /* clean up */
    free(filebuf);
    if (outfile != NULL) {
	fclose(output);
    }
    close(source);
    if (rval == 0)
	return (1);
    else
	return (0);
}
