/* coreography, a tool for examining ELF based memory images.
 * Copyright (C) 2003  John Ventura ventura@engination.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <elf.h>
#include <ctype.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include "cmd.h"
#include "addresses.h"

/*    This function allows the user to run commands on a specified
 *    chunk of memory.  Instead of specifying the pointer to
 *    that chunk, this function looks at the "linking view"
 *    and goes by section headers.  If the section name in
 *    strtab matches what the user wants, then go with that
 *    section.
 */

int checktextsanity(uint8_t * buf, int len);
int strings(uint8_t * buf, int len, uint32_t address, FILE * desc);
int segdumpascii(uint8_t * buf, int len, uint32_t address, FILE * desc);
int segdumpbinary(uint8_t * buf, int len, FILE * desc);
int search(uint8_t * buf, int buflen, uint32_t address, uint8_t * query,
	   int querylen, FILE * desc);

int selectbyname(uint8_t * buf, int buflen, uint32_t address, uint32_t len,
		 int binary, int cmdval, uint8_t * query, int querylen,
		 char *secname, FILE * desc)
{
    int rval = 0;
    int i;
    int textsanity;
    int memsize = 0;
    char *strtab = NULL;
    uint8_t *startbuf = NULL;
    uint32_t vaddr;
    Elf32_Ehdr *elfheader;
    Elf32_Phdr *programheader;
    Elf32_Shdr *sectionheader;

    elfheader = (Elf32_Ehdr *) buf;
    programheader = (Elf32_Phdr *) (buf + elfheader->e_phoff);
    sectionheader = (Elf32_Shdr *) (buf + elfheader->e_shoff);

    /* If strtab is not within the buffer, don't try to read it.
     */
    if (sectionheader[elfheader->e_shstrndx].sh_offset >= buflen) {
	fprintf(stderr, "can't find strtab\n");
	exit(1);
    }
    /* make sure data in strtab is printable.
     * if we're in the wrong section, that could be bad.
     */
    textsanity = checktextsanity((buf +
				  sectionheader[elfheader->
						e_shstrndx].
				  sh_offset),
				 sectionheader[elfheader->
					       e_shstrndx].sh_size);

    if (!textsanity) {
	fprintf(stderr, "strtab contains invalid data\n");
	exit(1);
    }
    strtab =
	(char *) (buf + sectionheader[elfheader->e_shstrndx].sh_offset);

    for (i = 0; i < elfheader->e_shnum; i++) {
	/* if the specified name is the same as the section name,
	 * start the appropriate command with the right values.
	 */
	if (!strcmp((strtab + sectionheader[i].sh_name), secname)) {
	    /* If the section is defined as extending past the buffer, 
	     * don't continue.
	     */
	    if ((sectionheader[i].sh_offset + sectionheader[i].sh_size) >
		buflen) {
		fprintf(stderr, "values are not in range\n");
		exit(1);
	    }
	    startbuf = (buf + sectionheader[i].sh_offset);
	    memsize = sectionheader[i].sh_size;
	    vaddr = sectionheader[i].sh_addr;

	    switch (cmdval) {
	    case CMD_STRINGS:
		rval += strings(startbuf, memsize, vaddr, desc);
		break;
	    case CMD_SEGDUMP:
		if (binary)
		    rval += segdumpbinary(startbuf, memsize, desc);
		else
		    rval += segdumpascii(startbuf, memsize, vaddr, desc);
		break;
	    case CMD_SEARCH:
		rval +=
		    search(startbuf, memsize, vaddr, query, querylen,
			   desc);
		break;
	    }
	}
    }


    return (rval);
}
