/* coreography, a tool for examining ELF based memory images.
 * Copyright (C) 2003  John Ventura ventura@engination.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <elf.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include "cmd.h"
#include "addresses.h"

int strings(uint8_t * buf, int len, uint32_t address, FILE * desc);
int segdumpascii(uint8_t * buf, int len, uint32_t address, FILE * desc);
int segdumpbinary(uint8_t * buf, int len, FILE * desc);
int search(uint8_t * buf, int buflen, uint32_t address, uint8_t * query,
	   int querylen, FILE * desc);
uint8_t *getprocseg(int pid, uint32_t addr, int len);

/* inrange() returns true if the highest byte of a 32bit int (x) is of
 *  * a specified value (y).
 */
#define inrange(x,y,z) ((x >> z) == (y))
#define ischar(c)((c > 0x20) && (c < 0x7e))
/* yes, there are macros like ischar() in ctype.h, but I don't like them */

/* This function tries to find the segments that the user is requesting,
 * and then, it sends that segment (maybe more than one)  to the 
 * appropriate function.
 */

int selectbytag(uint8_t * buf, int len, int segmentmask, int binary,
		int cmdval, uint8_t * query, int querylen,
		int usepid, int pid, FILE * desc)
{
    int rval = 0;
    int i;
    Elf32_Ehdr *elfheader;
    Elf32_Phdr *programheader;
    uint8_t *startbuf = buf;
    int memsize = len;
    uint32_t vaddr = 0;
    int dosomething = 0;
    char procfilename[512];
    char procbuf[512];
    FILE *source;

    if (usepid) {
	snprintf(procfilename, sizeof(procfilename), "/proc/%i/maps", pid);
	source = fopen(procfilename, "rb");
	if (source == NULL) {
	    fprintf(stderr, "can't get process info\n");
	    exit(1);
	}
	while (fgets(procbuf, sizeof(procbuf), source)) {

	    vaddr =
		(uint32_t) strtoul((char *) strtok(procbuf, "- "), NULL,
				   16);
	    memsize =
		strtoul((char *) strtok(NULL, "- "), NULL, 16) - vaddr;
	    if (segmentmask == 0)
		dosomething++;
	    if ((segmentmask & MASK_TEXT)
		&& inrange(vaddr, START_TEXT_LINUX_BYTE, 24))
		dosomething++;
	    if ((segmentmask & MASK_STACK)
		&& inrange(vaddr, START_STACK_LINUX_BYTE, 24))
		dosomething++;
	    if ((segmentmask & MASK_MMAP)
		&& inrange(vaddr, START_MMAP_LINUX_BYTE, 28))
		dosomething++;

	    if (dosomething) {
		startbuf = (uint8_t *) getprocseg(pid, vaddr, memsize);
		switch (cmdval) {
		case CMD_STRINGS:
		    rval += strings(startbuf, memsize, vaddr, desc);
		    break;
		case CMD_SEGDUMP:
		    if (binary)
			rval += segdumpbinary(startbuf, memsize, desc);
		    else
			rval +=
			    segdumpascii(startbuf, memsize, vaddr, desc);
		    break;
		case CMD_SEARCH:
		    rval +=
			search(startbuf, memsize, vaddr, query, querylen,
			       desc);
		    break;
		}
		dosomething = 0;
		free(startbuf);
	    }
	}
	fclose(source);
    } else {
	elfheader = (Elf32_Ehdr *) buf;
	programheader = (Elf32_Phdr *) (buf + elfheader->e_phoff);

	/* This part goes through the section headers and tries to
	 * figure out which section the user is asking for.
	 */
	for (i = 0; i < elfheader->e_phnum; i++) {
	    /*show everything if nothing is specified */
	    if ((segmentmask == 0) && (programheader[i].p_filesz != 0)) {
		startbuf = (buf + programheader[i].p_offset);
		memsize = programheader[i].p_memsz;
		vaddr = programheader[i].p_vaddr;
		dosomething++;
	    }
	    if (segmentmask & MASK_TEXT) {
		if ((programheader[i].p_vaddr == START_TEXT_LINUX)
		    && (programheader[i].p_flags & PF_X)) {
		    startbuf = (buf + programheader[i].p_offset);
		    memsize = programheader[i].p_memsz;
		    vaddr = programheader[i].p_vaddr;
		    dosomething++;
		}
	    }
	    if (segmentmask & MASK_AUXILIARY) {
		if (programheader[i].p_type == PT_NOTE) {
		    startbuf = (buf + programheader[i].p_offset);
		    memsize = programheader[i].p_filesz;
		    vaddr = programheader[i].p_vaddr;
		    dosomething++;
		}
		/* If the user asked for text, and the memory segment is at the
		 * right address, and it's executable, then show it.
		 */
	    }
	    if (segmentmask & MASK_MMAP) {
		if (inrange
		    (programheader[i].p_vaddr, START_MMAP_LINUX_BYTE, 28)
		    && (programheader[i].p_filesz != 0)) {
		    startbuf = (buf + programheader[i].p_offset);
		    memsize = programheader[i].p_memsz;
		    vaddr = programheader[i].p_vaddr;
		    dosomething++;
		}
	    }
	    if (segmentmask & MASK_STACK) {
		if (programheader[i].p_vaddr + programheader[i].p_memsz ==
		    END_STACK_LINUX) {
		    startbuf = (buf + programheader[i].p_offset);
		    memsize = programheader[i].p_memsz;
		    vaddr = programheader[i].p_vaddr;
		    dosomething++;
		}
	    }
	    if (dosomething) {
		switch (cmdval) {
		case CMD_STRINGS:
		    rval += strings(startbuf, memsize, vaddr, desc);
		    break;
		case CMD_SEGDUMP:
		    if (binary)
			rval += segdumpbinary(startbuf, memsize, desc);
		    else
			rval +=
			    segdumpascii(startbuf, memsize, vaddr, desc);
		    break;
		case CMD_SEARCH:
		    rval +=
			search(startbuf, memsize, vaddr, query, querylen,
			       desc);
		    break;
		}
		dosomething = 0;
	    }
	}
    }
    return (rval);
}
